      subroutine stadj(uadj, vadj, phiadj, u, v, phi, uobs, vobs, &
                       phiobs, UObsStddev, VObsStddev, PhiObsStddev, n, m, nt)
!***********************************************************
!  in this subroutine the adjoint variables have the same
!  name  as the direct variables in the direct system
!  the adjoint code is derived by writing backward
!  the linear tangent model
!  the adjoint of each instruction is written step by step
!********************************************************
      implicit none

! size parameters
      integer, intent(in) :: n, m, nt

! state variables
      double precision, intent(in), dimension(n, m, nt) :: u, v, phi

! observation variables
      double precision, intent(in), dimension(n, m, nt) :: uobs, vobs, &
                                                             phiobs

! weighting variables
      double precision, intent(in) :: UObsStddev, VObsStddev, PhiObsStddev

! adjoint variables
      double precision, intent(out), dimension(n, m, nt) :: uadj, &
                                                              vadj, &
                                                              phiadj

! domain length and depth in m
      double precision :: length, depth, h0, h1, h2

! Namelist for initial condition parameter data
      namelist /icdata/ length, depth, h0, h1, h2

! local variables
      integer :: im1, ip1, jm1, jp1, nm1, i, j, k, kp1, kpass, mm1, ntm1
      double precision :: htemp, ztemp, dt, dx, dy, hy, hhx, hx, hhy, &
                          dphival, dvval, duval, uval, vval, phival, &
                          udt, vdt, phidt

! Coriolis in 1/s
      double precision :: f
      data f/1.d-4/

! read IC parameters from namelist
      open(unit=42,file="namelist.swm",status='old',form='formatted')
      read(42,icdata)
      close(42)

      uadj = 0.
      vadj = 0.
      phiadj = 0.

! finite difference variables
      dx = (2.*length)/(n-1)
      dy = (2.*depth)/(m-1)
      hx = 1./(4.*dx)
      hy = 1./(4.*dy)
      hhx = 2.*hx
      hhy = 2.*hy

      ztemp= 300.
      htemp= 2.*ztemp

      ntm1=nt-1
      nm1 = n - 1
      mm1 = m - 1

      do k = ntm1, 0, -1
          kp1=k+1

          do j=2,m-1
              do i=2,n-1
                  call obs_operator(u(i,j,kp1),v(i,j,kp1),&
                                    phi(i,j,kp1),uval, &
                                    vval,phival)

                  call obs_operator_adj(u(i,j,kp1),v(i,j,kp1), &
                                        phi(i,j,kp1),duval, &
                                        dvval,dphival)

                  uadj(i,j,kp1)   = uadj(i,j,kp1) + &
                                    2.*(uval-uobs(i,j,kp1))*duval/(UObsStddev**2)

                  vadj(i,j,kp1)   = vadj(i,j,kp1) + &
                                    2.*(vval-vobs(i,j,kp1))*dvval/(VObsStddev**2)

                  phiadj(i,j,kp1) = phiadj(i,j,kp1) &
                                  + 2.*(phival-phiobs(i,j,kp1))*dphival/(PhiObsStddev**2)

              enddo
          enddo

          do i=1,n
!             bottom boundary
              phiadj(i,2,kp1) = phiadj(i,2,kp1) + phiadj(i,1,kp1)
              uadj(i,2,kp1)   = uadj(i,2,kp1)   + uadj(i,1,kp1)
          enddo

          do i=2,mm1
!             right boundary
              phiadj(2,i,kp1) = phiadj(2,i,kp1) + phiadj(n,i,kp1)
              vadj(2,i,kp1)   = vadj(2,i,kp1) +   vadj(n,i,kp1)
              uadj(2,i,kp1)   = uadj(2,i,kp1) +   uadj(n,i,kp1)
          enddo

          do i=1,n
!             top boundary
              phiadj(i,mm1,kp1) = phiadj(i,mm1,kp1) + phiadj(i,m,kp1)
              uadj(i,mm1,kp1)   = uadj(i,mm1,kp1)   + uadj(i,m,kp1)
          enddo

          do i=2,mm1
!             left boundary
              phiadj(nm1,i,kp1) = phiadj(nm1,i,kp1) + phiadj(1,i,kp1)
              vadj(nm1,i,kp1)   = vadj(nm1,i,kp1)   + vadj(1,i,kp1)
              uadj(nm1,i,kp1)   = uadj(nm1,i,kp1)   + uadj(1,i,kp1)
          enddo

          if (k .gt. 0) then
              if(k .eq. 1) then
                  kpass = k
                  dt = ztemp
              else
                  kpass = k-1
                  dt = htemp
              endif

              do j = 2, mm1
                  do i = 2, nm1
                      ip1 = i + 1
                      im1 = i - 1
                      jp1 = j + 1
                      jm1 = j - 1

! 1
                      phiadj(i,j,kpass) = phiadj(i,j,kpass) + phiadj(i,j,kp1)
                      phidt = -dt*phiadj(i,j,kp1)
! 2
                      vadj(i,j,kpass) = vadj(i,j,kpass) + vadj(i,j,kp1)
                      vdt = -dt*vadj(i,j,kp1)
! 3
                      uadj(i,j,kpass) = uadj(i,j,kpass) + uadj(i,j,kp1)
                      udt = -dt*uadj(i,j,kp1)
! 4
                      phiadj(i,jp1,k) = phiadj(i,jp1,k) + phidt*v(i,jp1,k)*hhy
                      phiadj(i,jm1,k) = phiadj(i,jm1,k) - phidt*v(i,jm1,k)*hhy
! 5 (update du 24 04 89)
                      phiadj(ip1,j,k) = phiadj(ip1,j,k) + phidt*u(ip1,j,k)*hhx
                      phiadj(im1,j,k) = phiadj(im1,j,k) - phidt*u(im1,j,k)*hhx
! 10
                      phiadj(i,jp1,k) = phiadj(i,jp1,k) + vdt*hhy
                      phiadj(i,jm1,k) = phiadj(i,jm1,k) - vdt*hhy
! 11
                      phiadj(ip1,j,k) = phiadj(ip1,j,k) + udt*hhx
                      phiadj(im1,j,k) = phiadj(im1,j,k) - udt*hhx
! 6
                      vadj(i,jp1,k) = vadj(i,jp1,k) + phidt*phi(i,jp1,k)*hhy
                      vadj(i,jm1,k) = vadj(i,jm1,k) - phidt*phi(i,jm1,k)*hhy
! 12  (update le 26 01 90)
                      vadj(i,jp1,k) = vadj(i,jp1,k) + vdt*(v(i,jp1,k) - v(i,jm1,k))*hy
                      vadj(i,jm1,k) = vadj(i,jm1,k) + vdt*(v(i,jp1,k) - v(i,jm1,k))*hy
! 14 (update le 26 01 90)
                      vadj(i,jp1,k) = vadj(i,jp1,k) + udt*(u(i,jp1,k) - u(i,jm1,k))*hy
                      vadj(i,jm1,k) = vadj(i,jm1,k) + udt*(u(i,jp1,k) - u(i,jm1,k))*hy
! 16 (update le 26 01 90)
                      vadj(i,jp1,k) = vadj(i,jp1,k) + vdt*(v(i,jp1,k) + v(i,jm1,k))*hy
                      vadj(i,jm1,k) = vadj(i,jm1,k) - vdt*(v(i,jp1,k) + v(i,jm1,k))*hy
! 17  (update le 26 01 90)
                      vadj(ip1,j,k) = vadj(ip1,j,k) + vdt*(u(ip1,j,k) + u(im1,j,k))*hx
                      vadj(im1,j,k) = vadj(im1,j,k) - vdt*(u(ip1,j,k) + u(im1,j,k))*hx
! 7
                      uadj(ip1,j,k) = uadj(ip1,j,k) + phidt*phi(ip1,j,k)*hhx
                      uadj(im1,j,k) = uadj(im1,j,k) - phidt*phi(im1,j,k)*hhx
! 13  (update le 26 01 90)
                      uadj(ip1,j,k) = uadj(ip1,j,k) + vdt*(v(ip1,j,k) - v(im1,j,k))*hx
                      uadj(im1,j,k) = uadj(im1,j,k) + vdt*(v(ip1,j,k) - v(im1,j,k))*hx
! 15 (update le 26 01 90)
                      uadj(ip1,j,k) = uadj(ip1,j,k) + udt*(u(ip1,j,k) - u(im1,j,k))*hx
                      uadj(im1,j,k) = uadj(im1,j,k) + udt*(u(ip1,j,k) - u(im1,j,k))*hx
! 18  (update le 26 01 90)
                      uadj(i,jp1,k) = uadj(i,jp1,k) + udt*(v(i,jp1,k) + v(i,jm1,k))*hy
                      uadj(i,jm1,k) = uadj(i,jm1,k) - udt*(v(i,jp1,k) + v(i,jm1,k))*hy
! 19  (update le 26 01 90)
                      uadj(ip1,j,k) = uadj(ip1,j,k) + udt*(u(ip1,j,k) + u(im1,j,k))*hx
                      uadj(im1,j,k) = uadj(im1,j,k) - udt*(u(ip1,j,k) + u(im1,j,k))*hx

! 8
                      vadj(i,j,k) = vadj(i,j,k) - udt*f
! 9
                      uadj(i,j,k) = uadj(i,j,k) + vdt*f
                  enddo
              enddo
          endif
      enddo

      return
      end
