      subroutine cg_descent (grad_tol, x, dim, cg_value, cg_grad,&
                       status, gnorm, f, iter, nfunc, ngrad,&
                       d, g, xtemp, gtemp)

      double precision x (*), d (*), g (*), xtemp (*), gtemp (*),&
                 delta, sigma, eps,&
                 gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                 wolfe_hi, wolfe_lo, awolfe_hi,&
                 QuadCutOff, StopFac, AWolfeFac,&
                 zero, feps, psi0, psi1, psi2,&
                 grad_tol, delta2, eta_sq, Qk,&
                 f, ftemp, gnorm, xnorm, gnorm2, dnorm2, denom,&
                 t, t1, t2, t3, t4, dphi, dphi0, alpha, talpha,&
                 yk, yk2, ykgk, dkyk, beta

      integer          n, n5, n6, nf, ng, info, nrestart,&
                 nexpand, nsecant, maxit,&
                 iter, status, nfunc, ngrad,&
                 i, j, i1, i2, i3, i4, dim

      logical    PertRule, QuadOK, QuadStep, PrintLevel,&
                 PrintFinal, StopRule, AWolfe, Step, debug,&
                 cg_tol

      external   cg_value, cg_grad

      common /cgparms/delta, sigma, eps,&
                gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                wolfe_hi, wolfe_lo, awolfe_hi,&
                QuadCutOff, StopFac, AWolfeFac,&
                zero, feps, psi0, psi1, psi2,&
                n, n5, n6, nf, ng, info,&
                nrestart, nexpand, nsecant, maxit,&
                PertRule, QuadOK, QuadStep, PrintLevel,&
                PrintFinal, StopRule, AWolfe, Step, debug

! initialize the parameters

      call cg_init (grad_tol, dim)

      if ( Step ) then
          alpha = gnorm
      endif
      delta2 = 2*delta - 1
      eta_sq = eta*eta
      iter = 0
      Ck = 0
      Qk = 0

! initial function and gradient evaluations, initial direction

      call cg_value (f, x, n)
      nf = nf + 1
       print*,'in cg f=',f
       write(*,*) 'in cg f=',f

      call cg_grad (g, x, n)
      ng = ng + 1
      f0 = f + f
      gnorm = zero
      xnorm = zero
      gnorm2 = zero
      do i = 1, n5
          xnorm = max1 (xnorm, abs (x (i)))
          t = g (i)
          d (i) = -t
          gnorm = max (gnorm, abs(t))
          gnorm2 = gnorm2 + t*t
      enddo
      do i = n6, n, 5
          xnorm = max (xnorm, abs (x (i)))
          t = g (i)
          gnorm = max (gnorm, abs (t))
          d (i)   = -t
          j = i + 1
          t1 = g (j)
          d (j) = -t1
          gnorm = max (gnorm, abs (t1))
          xnorm = max (xnorm, abs (x (j)))
          j = i + 2
          t2 = g (j)
          d (j) = -t2
          gnorm = max (gnorm, abs (t2))
          xnorm = max (xnorm, abs (x (j)))
          j = i + 3
          t3 = g (j)
          d (j) = -t3
          gnorm = max (gnorm, abs (t3))
          xnorm = max (xnorm, abs (x (j)))
          j = i + 4
          t4 = g (j)
          d (j) = -t4
          gnorm = max (gnorm, abs (t4))
          xnorm = max (xnorm, abs (x (j)))
          gnorm2 = gnorm2 + t*t + t1*t1 + t2*t2 + t3*t3 + t4*t4
      enddo

      if ( StopRule ) then
          tol = max (gnorm*StopFac, tol)
      endif

      if ( PrintLevel ) then
          write (*, 10) iter, f, gnorm, max(nf,ng), AWolfe
          write (84, *) max(nf,ng), f, gnorm
10        format ('iter: ', i5, ' f= ', e14.6,&
            ' gnorm= ', e14.6, ' nfunc = ', i5, ' AWolfe= ', l2)
      endif

      if ( cg_tol (f, gnorm) ) goto 100

      dphi0 = -gnorm2
      if ( .not.Step ) then
          alpha = psi0*xnorm/gnorm
          if ( xnorm .eq. zero ) then
              if ( f .ne. zero ) then
                  alpha = psi0*abs (f)/gnorm2
              else
                  alpha = 1.d0
              endif
          endif
      endif

! start the conjugate gradient iteration

      do iter = 1, maxit
          QuadOK = .false.
          alpha = psi2*alpha
          if ( QuadStep ) then
              if ( f .ne. zero ) then
                  t = abs ((f-f0)/f)
              else
                  t = 1.d0
              endif
              if ( t .gt. QuadCutOff ) then
                  talpha = psi1*alpha
                  call cg_step (xtemp, x, d, talpha)
                  call cg_value (ftemp, xtemp, n)
                  nf = nf + 1
                  if ( ftemp .lt. f ) then
                     denom = 2.0d0*(((ftemp-f)/talpha)-dphi0)
                     if ( denom .gt. zero ) then
                         QuadOK = .true.
                         alpha = -dphi0*talpha/denom
                     endif
                  endif
              endif
          endif
          f0 = f

!          if ( PrintLevel ) then
!              write (*, 20) QuadOK, alpha, f0, dphi0
!20            format ('QuadOK:', l2, ' initial a:',&
!                 e14.6,' f0:', e14.6, ' dphi', e14.6)
!          endif

! parameters in Wolfe and approximiate Wolfe conditions, and in update

          Qk = Qdecay*Qk + 1.
          Ck = Ck + (abs (f) - Ck)/Qk

          if ( PertRule ) then
              fpert = f + eps*Ck
          else
              fpert = f + eps
          endif

          wolfe_hi = delta*dphi0
          wolfe_lo = sigma*dphi0
          awolfe_hi = delta2*dphi0
          if ( AWolfe ) then
              call cg_line  (alpha, f, dphi, dphi0, x, xtemp, d, gtemp,&
                     cg_value, cg_grad)
          else
              call cg_lineW (alpha, f, dphi, dphi0, x, xtemp, d, gtemp,&
                     cg_value, cg_grad)
          endif

          if ( info .gt. 0 ) goto 100
!
! Test for convergence to within machine epsilon
!
          if ( -alpha*dphi0 .le. feps*abs (f) ) then
              info = 1
              goto 100
          endif

! compute beta, yk2, gnorm, gnorm2, dnorm2, update x and g,

          if ( mod (iter, nrestart) .ne. 0 ) then
              gnorm = zero
              dnorm2 = zero
              yk2 = zero
              ykgk = zero
              do i = 1, n5
                  x (i) = xtemp (i)
                  t = gtemp (i)
                  yk = t - g (i)
                  yk2 = yk2 + yk**2
                  ykgk = ykgk + yk*t
                  g (i) = t
                  gnorm = max (gnorm, abs (t))
                  dnorm2 = dnorm2 + d (i)**2
              enddo
              do i = n6, n, 5
                  x (i) = xtemp (i)
                  t = gtemp (i)
                  yk = t - g (i)
                  yk2 = yk2 + yk**2
                  ykgk = ykgk + yk*t
                  i1 = i + 1
                  x (i1) = xtemp (i1)
                  t1 = gtemp (i1)
                  i2 = i + 2
                  x (i2) = xtemp (i2)
                  t2 = gtemp (i2)
                  i3 = i + 3
                  x (i3) = xtemp (i3)
                  t3 = gtemp (i3)
                  i4 = i + 4
                  x (i4) = xtemp (i4)
                  t4 = gtemp (i4)
                  yk2 = yk2 + (t1-g (i1))**2 + (t2-g (i2))**2&
                      + (t3-g (i3))**2 + (t4-g (i4))**2
                  ykgk = ykgk + (t1-g (i1))*t1 + (t2-g (i2))*t2&
                        + (t3-g (i3))*t3 + (t4-g (i4))*t4
                  g (i) = t
                  gnorm = max (gnorm, abs (t))
                  g (i1) = t1
                  gnorm = max (gnorm, abs (t1))
                  g (i2) = t2
                  gnorm = max (gnorm, abs (t2))
                  g (i3) = t3
                  gnorm = max (gnorm, abs (t3))
                  g (i4) = t4
                  gnorm = max (gnorm, abs (t4))
                  dnorm2 = dnorm2 + d (i)**2 + d (i1)**2 + d (i2)**2&
                                       + d (i3)**2 + d (i4)**2
              enddo
              if ( cg_tol (f, gnorm) ) goto 100
              dkyk = dphi - dphi0
              beta = (ykgk - 2.d0*dphi*yk2/dkyk)/dkyk


              beta = max (beta,&
               -1.0/sqrt (min (eta_sq, gnorm2)*dnorm2))

!     update search direction d = -g + beta*dold

              gnorm2 = zero
              do i = 1, n5
                  t = g (i)
                  d (i) = -t + beta*d (i)
                  gnorm2 = gnorm2 + t*t
              enddo
              do i = n6, n, 5
                  d (i) = -g (i) + beta*d (i)
                  i1 = i + 1
                  d (i1) = -g (i1) + beta*d (i1)
                  i2 = i + 2
                  d (i2) = -g (i2) + beta*d (i2)
                  i3 = i + 3
                  d (i3) = -g (i3) + beta*d (i3)
                  i4 = i + 4
                  d (i4) = -g (i4) + beta*d (i4)
                  gnorm2 = gnorm2 + g (i)**2 + g (i1)**2 + g (i2)**2&
                                       + g (i3)**2 + g (i4)**2
              enddo
              dphi0 = -gnorm2 + beta*dphi

          else

!     search direction d = -g

!              if ( PrintLevel ) then
!                  write (*, *) "RESTART CG"
!              endif
              gnorm = zero
              gnorm2 = zero
              do i = 1, n5
                  x (i) = xtemp (i)
                  t = gtemp (i)
                  g (i) = t
                  d (i) = -t
                  gnorm = max (gnorm, abs(t))
                  gnorm2 = gnorm2 + t*t
              enddo
              do i = n6, n, 5
                  x (i) = xtemp (i)
                  t = gtemp (i)
                  g (i) = t
                  d (i) = -t
                  gnorm = max (gnorm, abs(t))
                  j = i + 1
                  x (j) = xtemp (j)
                  t1 = gtemp (j)
                  g (j) = t1
                  d (j) = -t1
                  gnorm = max (gnorm, abs(t1))
                  j = i + 2
                  x (j) = xtemp (j)
                  t2 = gtemp (j)
                  g (j) = t2
                  d (j) = -t2
                  gnorm = max (gnorm, abs(t2))
                  j = i + 3
                  x (j) = xtemp (j)
                  t3 = gtemp (j)
                  g (j) = t3
                  d (j) = -t3
                  gnorm = max (gnorm, abs(t3))
                  j = i + 4
                  x (j) = xtemp (j)
                  t4 = gtemp (j)
                  g (j) = t4
                  d (j) = -t4
                  gnorm = max (gnorm, abs(t4))
                  gnorm2 = gnorm2 + t*t + t1*t1 + t2*t2 + t3*t3 + t4*t4
              enddo
              if ( cg_tol (f, gnorm) ) goto 100
              dphi0 = -gnorm2
          endif
          if ( .not.AWolfe ) then
              if ( abs (f-f0) .lt. AWolfeFac*Ck ) then
                  AWolfe = .true.
              endif
          endif

          if ( PrintLevel ) then
              write (84, *) max(nf,ng), f, gnorm
              write (*, 10) iter, f, gnorm, max(nf, ng), AWolfe
          endif

          if ( debug ) then
              if ( f .gt. f0 + 1.e-10*Ck ) then
                  write (*, 270)
                  write (*, 50) f, f0
50                format (' new value:', e30.16, 'old value:', e30.16)
                  stop
              endif
          endif

          if ( dphi0 .gt. zero ) then
             info = 5
             goto 100
          endif
      enddo
      info = 2
100   nfunc = nf
      write(*,*) 'in cg 458 info=',info
      ngrad = ng
      status = info
      if ( info .gt. 2 ) then
          gnorm = zero
          do i = 1, n
              x (i) = xtemp (i)
              g (i) = gtemp (i)
              gnorm = max (gnorm, abs(g (i)))
          enddo
      endif
      write(*,*) '466 printfinal',printfinal
      if ( PrintFinal ) then
          write (*, *) 'Termination status:', status
          if ( status .eq. 0 ) then
              write (*, 200)
          else if ( status .eq. 1 ) then
              write (*, 210)
          else if ( status .eq. 2 ) then
              write (*, 220) maxit
              write (*, 300)
              write (*, 400) grad_tol
          else if ( status .eq. 3 ) then
              write (*, 230)
              write (*, 300)
              write (*, 430)
              write (*, 410)
          else if ( status .eq. 4 ) then
              write (*, 240)
              write (*, 300)
              write (*, 400) grad_tol
          else if ( status .eq. 5 ) then
              write (*, 250)
          else if ( status .eq. 6 ) then
              write (*, 260)
              write (*, 300)
              write (*, 400) grad_tol
              write (*, 410)
              write (*, 420)
          else if ( status .eq. 7 ) then
              write (*, 260)
              write (*, 300)
              write (*, 400) grad_tol
          else if ( status .eq. 8 ) then
              write (*, 260)
              write (*, 300)
              write (*, 400) grad_tol
              write (*, 410)
              write (*, 420)
          endif
          write (*, 500) gnorm
          write (*, *) 'function value:', f
          write (*, *) 'cg iterations:', iter
          write (*, *) 'function evaluations:', nfunc
          write (*, *) 'gradient evaluations:', ngrad
      endif
      return
200   format (' Convergence tolerance for gradient satisfied')
210   format (' Terminating since change in function value <= feps*|f|')
220   format (' Total number of iterations exceed max allow:', i10)
230   format (' Slope always negative in line search')
240   format (' Line search fails, too many secant steps')
250   format (' Search direction not a descent direction')
260   format (' Line search fails')
270   format (' Debugger is on, function value does not improve')
300   format (' Possible causes of this error message:')
400   format ('   - your tolerance (grad_tol = ', d10.4,&
        ') may be too strict')
410   format ('   - your gradient routine has an error')
420   format ('   - parameter epsilon in cg_descent.parm is too small')
430   format ('   - your cost function has an error')
500   format (' absolute largest component of gradient: ', d10.4)
      end

      subroutine cg_init (grad_tol, dim)

      double precision delta, sigma, eps,&
                 gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                 wolfe_hi, wolfe_lo, awolfe_hi,&
                 QuadCutOff, StopFac, AWolfeFac,&
                 zero, feps, psi0, psi1, psi2,&
                 grad_tol, restart_fac, maxit_fac

      integer          n, n5, n6, nf, ng, info, nrestart,&
                 nexpand, nsecant, maxit, dim

      logical          PertRule, QuadOK, QuadStep, PrintLevel,&
                 PrintFinal, StopRule, AWolfe, Step, debug

      common /cgparms/delta, sigma, eps,&
                gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                wolfe_hi, wolfe_lo, awolfe_hi,&
                QuadCutOff, StopFac, AWolfeFac,&
                zero, feps, psi0, psi1, psi2,&
                n, n5, n6, nf, ng, info,&
                nrestart, nexpand, nsecant, maxit,&
                PertRule, QuadOK, QuadStep, PrintLevel,&
                PrintFinal, StopRule, AWolfe, Step, debug

      n = dim
      tol = grad_tol
      open (10, file='cg_par')
      read (10, *) delta
      read (10, *) sigma
      read (10, *) eps
      read (10, *) gamma
      read (10, *) rho
      read (10, *) eta
      read (10, *) psi0
      read (10, *) psi1
      read (10, *) psi2
      read (10, *) QuadCutOff
      read (10, *) StopFac
      read (10, *) AWolfeFac
      read (10, *) restart_fac
      read (10, *) maxit_fac
      read (10, *) feps
      read (10, *) Qdecay
      read (10, *) nexpand
      read (10, *) nsecant
      read (10, *) PertRule
      read (10, *) QuadStep
      read (10, *) PrintLevel
      read (10, *) PrintFinal
      read (10, *) StopRule
      read (10, *) AWolfe
      read (10, *) Step
      read (10, *) debug
      nrestart = n*restart_fac
      maxit = n*maxit_fac
      zero = 0.0
      info = 0
      n5 = mod (n, 5)
      n6 = n5 + 1
      nf = 0
      ng = 0
      close (10)
      return
      end


      logical function cg_Wolfe (alpha, f, dphi)

      double precision delta, sigma, eps,&
                 gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                 wolfe_hi, wolfe_lo, awolfe_hi,&
                 QuadCutOff, StopFac, AWolfeFac,&
                 zero, feps, psi0, psi1, psi2,&
                 alpha, f, dphi

      integer          n, n5, n6, nf, ng, info, nrestart,&
                 nexpand, nsecant, maxit

      logical          PertRule, QuadOK, QuadStep, PrintLevel,&
                 PrintFinal, StopRule, AWolfe, Step, debug

      common /cgparms/delta, sigma, eps,&
                gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                wolfe_hi, wolfe_lo, awolfe_hi,&
                QuadCutOff, StopFac, AWolfeFac,&
                zero, feps, psi0, psi1, psi2,&
                n, n5, n6, nf, ng, info,&
                nrestart, nexpand, nsecant, maxit,&
                PertRule, QuadOK, QuadStep, PrintLevel,&
                PrintFinal, StopRule, AWolfe, Step, debug

      if ( dphi .ge. wolfe_lo ) then

! test original Wolfe conditions

          if ( f-f0 .le. alpha*wolfe_hi ) then
              cg_Wolfe = .true.
!              if ( PrintLevel ) then
!                  write (*, 10) f, f0, alpha*wolfe_hi, dphi
!10                format (' wolfe f:', e14.6, ' f0: ',&
!                    e14.6, e14.6, ' dphi:', e14.6)
!              endif
              return

! test approximate Wolfe conditions

          elseif ( AWolfe ) then
              if ( (f .le. fpert).and.(dphi .le. awolfe_hi) ) then
                  cg_Wolfe = .true.
!                  if ( PrintLevel ) then
!                      write (*, 20) f, fpert, dphi, awolfe_hi
!20                        format ('f:', e14.6, ' fpert:', e14.6,&
!                            ' dphi: ', e14.6, ' fappx:', e14.6)
!                  endif
                  return
              endif
          endif
      endif
      cg_Wolfe = .false.
      return
      end

! check for convergence of the cg iterations
!      (double) f       -- function value associated with stepsize
!      (double) gnorm   -- gradient (infinity) norm

      logical function cg_tol (f, gnorm)

      double precision delta, sigma, eps,&
                 gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                 wolfe_hi, wolfe_lo, awolfe_hi,&
                 QuadCutOff, StopFac, AWolfeFac,&
                 zero, feps, psi0, psi1, psi2,&
                 f, gnorm

      integer          n, n5, n6, nf, ng, info, nrestart,&
                 nexpand, nsecant, maxit

      logical          PertRule, QuadOK, QuadStep, PrintLevel,&
                 PrintFinal, StopRule, AWolfe, Step, debug

      common /cgparms/delta, sigma, eps,&
                gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                wolfe_hi, wolfe_lo, awolfe_hi,&
                QuadCutOff, StopFac, AWolfeFac,&
                zero, feps, psi0, psi1, psi2,&
                n, n5, n6, nf, ng, info,&
                nrestart, nexpand, nsecant, maxit,&
                PertRule, QuadOK, QuadStep, PrintLevel,&
                PrintFinal, StopRule, AWolfe, Step, debug

      if ( StopRule ) then
          if ( gnorm .le. tol ) then
              cg_tol = .true.
              return
          endif
      else
          if ( gnorm .le. tol*(1.0 + abs (f)) ) then
              cg_tol = .true.
              return
          endif
      endif
      cg_tol = .false.
      return
      end


      double precision function cg_dot (x, y)

      double precision delta, sigma, eps,&
                 gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                 wolfe_hi, wolfe_lo, awolfe_hi,&
                 QuadCutOff, StopFac, AWolfeFac,&
                 zero, feps, psi0, psi1, psi2,&
                 x (*), y(*), t

      integer          n, n5, n6, nf, ng, info, nrestart,&
                 nexpand, nsecant, maxit, i

      logical          PertRule, QuadOK, QuadStep, PrintLevel,&
                 PrintFinal, StopRule, AWolfe, Step, debug

      common /cgparms/delta, sigma, eps,&
                gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                wolfe_hi, wolfe_lo, awolfe_hi,&
                QuadCutOff, StopFac, AWolfeFac,&
                zero, feps, psi0, psi1, psi2,&
                n, n5, n6, nf, ng, info,&
                nrestart, nexpand, nsecant, maxit,&
                PertRule, QuadOK, QuadStep, PrintLevel,&
                PrintFinal, StopRule, AWolfe, Step, debug

      t = zero
      do i = 1, n5
          t = t + x (i)*y (i)
      enddo
      do i = n6, n, 5
          t = t + x (i)*y(i) + x (i+1)*y (i+1) + x (i+2)*y (i+2)&
                       + x (i+3)*y (i+3) + x (i+4)*y (i+4)
      enddo
      cg_dot = t
      return
      end



      subroutine cg_step (xtemp, x, d, alpha)

      double precision delta, sigma, eps,&
                 gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                 wolfe_hi, wolfe_lo, awolfe_hi,&
                 QuadCutOff, StopFac, AWolfeFac,&
                 zero, feps, psi0, psi1, psi2,&
                 xtemp (*), x (*), d (*), alpha

      integer          n, n5, n6, nf, ng, info, nrestart,&
                 nexpand, nsecant, maxit, i, j

      logical          PertRule, QuadOK, QuadStep, PrintLevel,&
                 PrintFinal, StopRule, AWolfe, Step, debug

      common /cgparms/delta, sigma, eps,&
                gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                wolfe_hi, wolfe_lo, awolfe_hi,&
                QuadCutOff, StopFac, AWolfeFac,&
                zero, feps, psi0, psi1, psi2,&
                n, n5, n6, nf, ng, info,&
                nrestart, nexpand, nsecant, maxit,&
                PertRule, QuadOK, QuadStep, PrintLevel,&
                PrintFinal, StopRule, AWolfe, Step, debug

      do i = 1, n5
          xtemp (i) = x(i) + alpha*d(i)
      enddo
      do i = n6, n, 5
          xtemp (i) = x (i) + alpha*d (i)
          j = i + 1
          xtemp (j) = x (j) + alpha*d (j)
          j = i + 2
          xtemp (j) = x (j) + alpha*d (j)
          j = i + 3
          xtemp (j) = x (j) + alpha*d (j)
          j = i + 4
          xtemp (j) = x (j) + alpha*d (j)
      enddo
      end



      subroutine cg_line (alpha, phi, dphi, dphi0, x, xtemp, d, gtemp,&
                    cg_value, cg_grad)

      double precision delta, sigma, eps,&
                 gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                 wolfe_hi, wolfe_lo, awolfe_hi,&
                 QuadCutOff, StopFac, AWolfeFac,&
                 zero, feps, psi0, psi1, psi2,&
                 x (*), xtemp (*), d (*), gtemp (*),&
                 a, dphia, b, dphib, alpha, phi, dphi, c,&
                 a0, da0, b0, db0, width, fquad, dphi0,&
                 cg_dot

      integer          n, n5, n6, nf, ng, info, nrestart,&
                 nexpand, nsecant, maxit,&
                 ngrow, nshrink, cg_update, iter, flag

      logical          PertRule, QuadOK, QuadStep, PrintLevel,&
                 PrintFinal, StopRule, AWolfe, Step, debug,&
                 cg_Wolfe

      external         cg_value, cg_grad

      common /cgparms/delta, sigma, eps,&
                gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                wolfe_hi, wolfe_lo, awolfe_hi,&
                QuadCutOff, StopFac, AWolfeFac,&
                zero, feps, psi0, psi1, psi2,&
                n, n5, n6, nf, ng, info,&
                nrestart, nexpand, nsecant, maxit,&
                PertRule, QuadOK, QuadStep, PrintLevel,&
                PrintFinal, StopRule, AWolfe, Step, debug

      call cg_step (xtemp, x, d, alpha)
      call cg_grad (gtemp, xtemp, n)
      ng = ng + 1
      dphi = cg_dot (gtemp, d)
!
!
      a = zero
      dphia = dphi0
      ngrow = 0
      nshrink = 0
      do while ( dphi .lt. zero )
          call cg_value (phi, xtemp, n)
          nf = nf + 1
!
!
          if ( QuadOK ) then
              if ( ngrow .eq. 0 ) fquad = min1 (phi, f0)
              if ( phi .le. fquad ) then
!                  if ( PrintLevel ) then
!                      write (*, 10) alpha, phi, fquad
!10                    format ('alpha:', e14.6, ' phi:', e14.6,&
!                        ' fquad:', e14.6)
!                  endif
                  if ( cg_Wolfe (alpha, phi, dphi) ) return
              endif
          endif
          if ( phi .le. fpert ) then
              a = alpha
              dphia = dphi
          else
!
! contraction phase
!
              b = alpha
              do while ( .true. )
                  alpha = .5d0*(a+b)
                  nshrink = nshrink + 1
                  if ( nshrink .gt. nexpand ) then
                      info = 6
                      return
                  endif
                  call cg_step (xtemp, x, d, alpha)
                  call cg_grad (gtemp, xtemp, n)
                  ng = ng + 1
                  dphi = cg_dot (gtemp, d)
                  if ( dphi .ge. zero ) goto 100
                  call cg_value (phi, xtemp, n)
                  nf = nf + 1
!                  if ( PrintLevel ) then
!                      write (*, 20) a, b, alpha, phi, dphi
!20                    format ('contract, a:', e14.6,&
!                        ' b:', e14.6, ' alpha:', e14.6,&
!                        ' phi:', e14.6, ' dphi:', e14.6)
!                  endif
                  if ( QuadOK .and. (phi .le. fquad) ) then
                      if ( cg_Wolfe (alpha, phi, dphi) ) return
                  endif
                  if ( phi .le. fpert ) then
                      a = alpha
                      dphia = dphi
                  else
                      b = alpha
                  endif
              enddo
          endif
!
! expansion phase
!
          ngrow = ngrow + 1
          if ( ngrow .gt. nexpand ) then
              info = 3
              return
          endif
          alpha = rho*alpha
          call cg_step (xtemp, x, d, alpha)
          call cg_grad (gtemp, xtemp, n)
          ng = ng + 1
          dphi = cg_dot (gtemp, d)
!          if ( PrintLevel ) then
!              write (*, 30) a, alpha, phi, dphi
!30            format ('expand,   a:', e14.6, ' alpha:', e14.6,&
!                 ' phi:', e14.6, ' dphi:', e14.6)
!          endif
      enddo
100   continue
      b = alpha
      dphib = dphi
      if ( QuadOK ) then
          call cg_value (phi, xtemp, n)
          nf = nf + 1
          if ( ngrow + nshrink .eq. 0 ) fquad = min1 (phi, f0)
          if ( phi .le. fquad ) then
              if ( cg_Wolfe (alpha, phi, dphi) ) return
          endif
      endif
      do iter = 1, nsecant
!          if ( PrintLevel ) then
!              write (*, 40) a, b, dphia, dphib
!40            format ('secant, a:', e14.6, ' b:', e14.6,&
!                 ' da:', e14.6, ' db:', e14.6)
!          endif
          width = gamma*(b - a)
          if ( -dphia .le. dphib ) then
              alpha = a - (a-b)*(dphia/(dphia-dphib))
          else
              alpha = b - (a-b)*(dphib/(dphia-dphib))
          endif
          c = alpha
          a0 = a
          b0 = b
          da0 = dphia
          db0 = dphib
          flag = cg_update (a, dphia, b, dphib, alpha, phi,&
              dphi, x, xtemp, d, gtemp, cg_value, cg_grad)
          if ( flag .gt. 0 ) then
              return
          else if ( flag .eq. 0 ) then
              if ( c .eq. a ) then
                  if ( dphi .gt. da0 ) then
                      alpha = c - (c-a0)*(dphi/(dphi-da0))
                  else
                      alpha = a
                  endif
              else
                  if ( dphi .lt. db0 ) then
                      alpha = c - (c-b0)*(dphi/(dphi-db0))
                  else
                      alpha = b
                  endif
              endif
              if ( (alpha .gt. a) .and. (alpha .lt. b) ) then
!                  if ( PrintLevel ) write (*, *) "2nd secant"
                  flag = cg_update (a, dphia, b, dphib, alpha, phi,&
                      dphi, x, xtemp, d, gtemp, cg_value, cg_grad)
                  if ( flag .gt. 0 ) return
              endif
          endif
!
!    bisection iteration
!
          if ( (b-a) .ge. width ) then
              alpha = .5d0*(b+a)
!              if ( PrintLevel ) write (*, *) "bisection"
              flag = cg_update (a, dphia, b, dphib, alpha, phi,&
                  dphi, x, xtemp, d, gtemp, cg_value, cg_grad)
              if ( flag .gt. 0 ) return
          else
              if ( b .le. a ) then
                  info = 7
                  return
              endif
          endif
      end do
      info = 4
      return
      end


      subroutine cg_lineW (alpha, phi, dphi, dphi0, x, xtemp, d, gtemp,&
                    cg_value, cg_grad)

      double precision delta, sigma, eps,&
                 gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                 wolfe_hi, wolfe_lo, awolfe_hi,&
                 QuadCutOff, StopFac, AWolfeFac,&
                 zero, feps, psi0, psi1, psi2,&
                 x (*), xtemp (*), d (*), gtemp (*),&
                 a, dpsia, b, dpsib, alpha, phi, dphi, c,&
                 a0, da0, b0, db0, width, fquad, dphi0,&
                 cg_dot, psi, dpsi

      integer          n, n5, n6, nf, ng, info, nrestart,&
                 nexpand, nsecant, maxit,&
                 ngrow, nshrink, cg_updateW, iter, flag

      logical          PertRule, QuadOK, QuadStep, PrintLevel,&
                 PrintFinal, StopRule, AWolfe, Step, debug,&
                 cg_Wolfe

      external         cg_value, cg_grad

      common /cgparms/delta, sigma, eps,&
                gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                wolfe_hi, wolfe_lo, awolfe_hi,&
                QuadCutOff, StopFac, AWolfeFac,&
                zero, feps, psi0, psi1, psi2,&
                n, n5, n6, nf, ng, info,&
                nrestart, nexpand, nsecant, maxit,&
                PertRule, QuadOK, QuadStep, PrintLevel,&
                PrintFinal, StopRule, AWolfe, Step, debug

      call cg_step (xtemp, x, d, alpha)
      call cg_grad (gtemp, xtemp, n)
      ng = ng + 1
      dphi = cg_dot (gtemp, d)
      dpsi = dphi - wolfe_hi
!
!
      a = zero
      dpsia = dphi0 - wolfe_hi
      ngrow = 0
      nshrink = 0
      do while ( dpsi .lt. zero )
          call cg_value (phi, xtemp, n)
          psi = phi - alpha*wolfe_hi

          nf = nf + 1
!
! if quadstep in effect and quadratic conditions hold, check wolfe condition
!
          if ( QuadOK ) then
              if ( ngrow .eq. 0 ) fquad = min1 (phi, f0)
              if ( phi .le. fquad ) then
!                  if ( PrintLevel ) then
!                      write (*, 10) alpha, phi, fquad
!10                    format ('alpha:', e14.6, ' phi:', e14.6,&
!                        ' fquad:', e14.6)
!                  endif
                  if ( cg_Wolfe (alpha, phi, dphi) ) return
              endif
          endif
          if ( psi .le. fpert ) then
              a = alpha
              dpsia = dpsi
          else
!
! contraction phase
!
              b = alpha
              do while ( .true. )
                  alpha = .5d0*(a+b)
                  nshrink = nshrink + 1
                  if ( nshrink .gt. nexpand ) then
                      info = 6
                      return
                  endif
                  call cg_step (xtemp, x, d, alpha)
                  call cg_grad (gtemp, xtemp, n)
                  ng = ng + 1
                  dphi = cg_dot (gtemp, d)
                  dpsi = dphi - wolfe_hi
                  if ( dpsi .ge. zero ) goto 100
                  call cg_value (phi, xtemp, n)
                  psi = phi - alpha*wolfe_hi
                  nf = nf + 1
!                  if ( PrintLevel ) then
!                      write (*, 20) a, b, alpha, phi, dphi
!20                    format ('contract, a:', e14.6,&
!                        ' b:', e14.6, ' alpha:', e14.6,&
!                        ' phi:', e14.6, ' dphi:', e14.6)
!                  endif
                  if ( QuadOK .and. (phi .le. fquad) ) then
                      if ( cg_Wolfe (alpha, phi, dphi) ) return
                  endif
                  if ( psi .le. fpert ) then
                      a = alpha
                      dpsia = dpsi
                  else
                      b = alpha
                  endif
              enddo
          endif
!
! expansion phase
!
          ngrow = ngrow + 1
          if ( ngrow .gt. nexpand ) then
              info = 3
              return
          endif
          alpha = rho*alpha
          call cg_step (xtemp, x, d, alpha)
          call cg_grad (gtemp, xtemp, n)
          ng = ng + 1
          dphi = cg_dot (gtemp, d)
          dpsi = dphi - wolfe_hi
!          if ( PrintLevel ) then
!              write (*, 30) a, alpha, phi, dphi
!30            format ('expand,   a:', e14.6, ' alpha:', e14.6,&
!                 ' phi:', e14.6, ' dphi:', e14.6)
!              write (*, *) "expand, alpha:", alpha, "dphi:", dphi
!          endif
      enddo
100   continue
      b = alpha
      dpsib = dpsi
      if ( QuadOK ) then
          call cg_value (phi, xtemp, n)
          nf = nf + 1
          if ( ngrow + nshrink .eq. 0 ) fquad = min1 (phi, f0)
          if ( phi .le. fquad ) then
              if ( cg_Wolfe (alpha, phi, dphi) ) return
          endif
      endif
      do iter = 1, nsecant
!          if ( PrintLevel ) then
!              write (*, 40) a, b, dpsia, dpsib
!40            format ('secant, a:', e14.6, ' b:', e14.6,&
!                 ' da:', e14.6, ' db:', e14.6)
!          endif
          width = gamma*(b - a)
          if ( -dpsia .le. dpsib ) then
              alpha = a - (a-b)*(dpsia/(dpsia-dpsib))
          else
              alpha = b - (a-b)*(dpsib/(dpsia-dpsib))
          endif
          c = alpha
          a0 = a
          b0 = b
          da0 = dpsia
          db0 = dpsib
          flag = cg_updateW (a, dpsia, b, dpsib, alpha,&
               phi, dphi, dpsi, x, xtemp, d, gtemp,&
               cg_value, cg_grad)
          if ( flag .gt. 0 ) then
              return
          else if ( flag .eq. 0 ) then
              if ( c .eq. a ) then
                  if ( dpsi .gt. da0 ) then
                      alpha = c - (c-a0)*(dpsi/(dpsi-da0))
                  else
                      alpha = a
                  endif
              else
                  if ( dpsi .lt. db0 ) then
                      alpha = c - (c-b0)*(dpsi/(dpsi-db0))
                  else
                      alpha = b
                  endif
              endif
              if ( (alpha .gt. a) .and. (alpha .lt. b) ) then
!                  if ( PrintLevel ) write (*, *) "2nd secant"
                  flag = cg_updateW (a, dpsia, b, dpsib, alpha,&
                       phi, dphi, dpsi, x, xtemp, d, gtemp,&
                       cg_value, cg_grad)
                  if ( flag .gt. 0 ) return
              endif
          endif
!
!    bisection iteration
!
          if ( (b-a) .ge. width ) then
              alpha = .5d0*(b+a)
!              if ( PrintLevel ) write (*, *) "bisection"
              flag = cg_updateW (a, dpsia, b, dpsib, alpha,&
                   phi, dphi, dpsi, x, xtemp, d, gtemp,&
                   cg_value, cg_grad)
              if ( flag .gt. 0 ) return
          else
              if ( b .le. a ) then
                  info = 7
                  return
              endif
          endif
      end do
      info = 4
      return
      end
!

      integer function cg_update (a, dphia, b, dphib, alpha, phi,&
                    dphi, x, xtemp, d, gtemp, cg_value, cg_grad)

      double precision delta, sigma, eps,&
                 gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                 wolfe_hi, wolfe_lo, awolfe_hi,&
                 QuadCutOff, StopFac, AWolfeFac,&
                 zero, feps, psi0, psi1, psi2,&
                 a, dphia, b, dphib, alpha, phi, dphi,&
                 x (*), xtemp (*), d (*), gtemp (*),&
                 cg_dot

      integer          n, n5, n6, nf, ng, info, nrestart,&
                 nexpand, nsecant, maxit,&
                 nshrink

      logical          PertRule, QuadOK, QuadStep, PrintLevel,&
                 PrintFinal, StopRule, AWolfe, Step, debug,&
                 cg_Wolfe

      external         cg_value, cg_grad

      common /cgparms/delta, sigma, eps,&
                gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                wolfe_hi, wolfe_lo, awolfe_hi,&
                QuadCutOff, StopFac, AWolfeFac,&
                zero, feps, psi0, psi1, psi2,&
                n, n5, n6, nf, ng, info,&
                nrestart, nexpand, nsecant, maxit,&
                PertRule, QuadOK, QuadStep, PrintLevel,&
                PrintFinal, StopRule, AWolfe, Step, debug

      call cg_step (xtemp, x, d, alpha)
      call cg_value (phi, xtemp, n)
      nf = nf + 1
      call cg_grad (gtemp, xtemp, n)
      ng = ng + 1
      dphi = cg_dot (gtemp, d)
!      if ( PrintLevel ) then
!          write (*, 10) alpha, phi, dphi
!10        format ('update alpha:', e14.6, ' phi:', e14.6,&
!            ' dphi:', e14.6)
!      endif
      cg_update = 0
      if ( cg_Wolfe (alpha, phi, dphi) ) then
          cg_update = 1
          goto 110
      endif
      if ( dphi .ge. zero ) then
          b = alpha
          dphib = dphi
          goto 110
      else
          if ( phi .le. fpert ) then
              a = alpha
              dphia = dphi
              goto 110
          endif
      endif
      nshrink = 0
      b = alpha
      do while ( .true. )
          alpha = .5d0*(a+b)
          nshrink = nshrink + 1
          if ( nshrink .gt. nexpand ) then
              info = 8
              cg_update = 1
              goto 110
          endif
          call cg_step (xtemp, x, d, alpha)
          call cg_grad (gtemp, xtemp, n)
          ng = ng + 1
          dphi = cg_dot (gtemp, d)
          call cg_value (phi, xtemp, n)
          nf = nf + 1
!          if ( PrintLevel ) then
!              write (*, 20) a, alpha, phi, dphi
!20            format ('contract, a:', e14.6, ' alpha:', e14.6,&
!                 ' phi:', e14.6, ' dphi:', e14.6)
!          endif
          if ( cg_Wolfe (alpha, phi, dphi) ) then
              cg_update = 1
              goto 110
          endif
          if ( dphi .ge. zero ) then
              b = alpha
              dphib = dphi
              goto 100
          endif
          if ( phi .le. fpert ) then
!              if ( PrintLevel ) then
!                  write (*, *) "update a:", alpha, "dphia:", dphi
!              endif
              a = alpha
              dphia = dphi
          else
              b = alpha
          endif
      enddo
100   continue
      cg_update = -1
110   continue
!      if ( PrintLevel ) then
!          write (*, 200) a, b, dphia, dphib, cg_update
!200       format ('UP a:', e14.6, ' b:', e14.6,&
!             ' da:', e14.6, ' db:', e14.6, ' up:', i2)
!      endif
      return
      end


      integer function cg_updateW (a, dpsia, b, dpsib, alpha, phi, dphi,&
                      dpsi, x, xtemp, d, gtemp, cg_value, cg_grad)

      double precision delta, sigma, eps,&
                 gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                 wolfe_hi, wolfe_lo, awolfe_hi,&
                 QuadCutOff, StopFac, AWolfeFac,&
                 zero, feps, psi0, psi1, psi2,&
                 a, dpsia, b, dpsib, alpha, phi, dphi,&
                 x (*), xtemp (*), d (*), gtemp (*),&
                 cg_dot, psi, dpsi

      integer          n, n5, n6, nf, ng, info, nrestart,&
                 nexpand, nsecant, maxit, nshrink

      logical          PertRule, QuadOK, QuadStep, PrintLevel,&
                 PrintFinal, StopRule, AWolfe, Step, debug,&
                 cg_Wolfe

      external         cg_value, cg_grad

      common /cgparms/delta, sigma, eps,&
                gamma, rho, tol, eta, fpert, f0, Ck, Qdecay,&
                wolfe_hi, wolfe_lo, awolfe_hi,&
                QuadCutOff, StopFac, AWolfeFac,&
                zero, feps, psi0, psi1, psi2,&
                n, n5, n6, nf, ng, info,&
                nrestart, nexpand, nsecant, maxit,&
                PertRule, QuadOK, QuadStep, PrintLevel,&
                PrintFinal, StopRule, AWolfe, Step, debug

      call cg_step (xtemp, x, d, alpha)
      call cg_value (phi, xtemp, n)
      psi = phi - alpha*wolfe_hi
      nf = nf + 1
      call cg_grad (gtemp, xtemp, n)
      ng = ng + 1
      dphi = cg_dot (gtemp, d)
      dpsi = dphi - wolfe_hi
!      if ( PrintLevel ) then
!          write (*, 10) alpha, psi, dpsi
!10        format ('update alpha:', e14.6, ' psi:', e14.6,&
!            ' dpsi:', e14.6)
!      endif
      cg_updateW = 0
      if ( cg_Wolfe (alpha, phi, dphi) ) then
          cg_updateW = 1
          goto 110
      endif
      if ( dpsi .ge. zero ) then
          b = alpha
          dpsib = dpsi
          goto 110
      else
          if ( psi .le. fpert ) then
              a = alpha
              dpsia = dpsi
              goto 110
          endif
      endif
      nshrink = 0
      b = alpha
      do while ( .true. )
          alpha = .5d0*(a+b)
          nshrink = nshrink + 1
          if ( nshrink .gt. nexpand ) then
              info = 8
              cg_updateW = 1
              goto 110
          endif
          call cg_step (xtemp, x, d, alpha)
          call cg_grad (gtemp, xtemp, n)
          ng = ng + 1
          dphi = cg_dot (gtemp, d)
          dpsi = dphi - wolfe_hi
          call cg_value (phi, xtemp, n)
          psi = phi - alpha*wolfe_hi
          nf = nf + 1
!          if ( PrintLevel ) then
!              write (*, 20) a, alpha, phi, dphi
!20            format ('contract, a:', e14.6, ' alpha:', e14.6,&
!                 ' phi:', e14.6, ' dphi:', e14.6)
!          endif
          if ( cg_Wolfe (alpha, phi, dphi) ) then
              cg_updateW = 1
              goto 110
          endif
          if ( dpsi .ge. zero ) then
              b = alpha
              dpsib = dpsi
              goto 100
          endif
          if ( psi .le. fpert ) then
!              if ( PrintLevel ) then
!                  write (*, *) "update a:", alpha, "dpsia:", dpsi
!              endif
              a = alpha
              dpsia = dpsi
          else
              b = alpha
          endif
      enddo
100   continue
      cg_updateW = -1
110   continue
!      if ( PrintLevel ) then
!          write (*, 200) a, b, dpsia, dpsib, cg_updateW
!200       format ('UP a:', e14.6, ' b:', e14.6,&
!             ' da:', e14.6, ' db:', e14.6, ' up:', i2)
!      endif
      return
      end
